=begin
(c) TIG 2010
Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.
THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
___________________________________________________________________________

  FreeRotate.rb >>> Plugins Folder
___________________________________________________________________________

Usage:

Make a Selection.
Pick Menu > Plugins > Free Rotate
  or type freerotate into Ruby Console
  or click on Toolbar 'Free Rotate' [activated from View > Toolbars].
The Selection can consist of any entities - it is temporarily Grouped 
during the Rotation.

You are then prompted to pick the 'Pivot' point.
Pick a point ]OR to set the Pivot-Point at the center of the Selection's 
bounding-box press the Ctrl-key].
The Pivot-Point is marked with a Black dot.

You are next prompted to pick the 'Axis' point.
A line is drawn from the Pivot-Point as you move the cursor.
Normal inferencing applies.
Pick a point [OR if the Pivot is not the Selection's bounding-box center you 
can set the Axis-Point at the center of the Selection's bounding-box by 
pressing the Ctrl-key].
The Axis-Point is now marked with a Magenta dot.

Now a temporary gizmo 'globe' appears, centered on the Pivot-Point 
with its North-Pole set vertically up.  
You can snap to it OR any other object in the model.
As you move the cursor the Selection will Free-Rotate so that the Axis-Point 
always points towards the cursor position.

The tool always starts in 'Free-Rotate' mode but you can toggle it to be in 
'Axial-Rotate' mode at any stage by pressing the 'Tab' key - it's probably 
easist to do this before picking the first two points.
'Tab' toggles between these two modes.
The 'Constrained' Axial-Rotate mode rotates the Selection around the Axis.
If you haven't already set the Pivot and/or Axis points before you go into 
this mode you will be prompted to do so first.
In this mode the gizmo becomes a 'disk', centered on the Pivot-Point, 
with its normal along the Axis.
You can snap to it or any other object in the model.
You are first prompted to pick a 'Handle' point.
Then pick a 'Snap' point - as you move the cursor the Selection will Rotate 
around the Axis following the cursor.  
Alternatively once the 'disk' has appeared you can type in an Angle into 
the VCB in degrees [-/+ve note that positive angles are counterclockwise] 
at any time when in the Axial-Rotate mode.
When you pick the Snap-Point or press <Enter> for the entered Angle the 
Selection with pause at that Rotation.
After typing an Angle you will need to re-pick a Handle-Point if you want to 
then Axial-Rotate by picked points.
However, you can continue typing in other Angles [which are taken as relative 
to the current postion] as desired.

Im either mode you can the press Enter or Double-Click to Confirm.

In either mode Esc will Exit the tool and Undo any changes.
Also in either mode if you start to use another Tool any changes are Undone.

The VCB-text gives you the current Mode and its Options...

___________________________________________________________________________
TO RECAP...

Make a Selection

Start the Tool...

Optionally use 'Tab' to toggle between 'Free-Rotate' and 'Axial-Rotate'.

Pick-1 = Pivot-Point
Pick-2 = Axis-Point

Next...

When in 'Free-Rotate' mode [globe-gizmo]:-

  Tab='Axial-Rotate'
  
  Left_Click/Enter/Double_Click=Confirm
    i.e. Use the current Rotation and Exit the tool.
      
  Esc/Exit=Cancel
    i.e. to Abort and Undo all changes at any stage...
    an 'Exit' occurs when you start to use another Tool.
        
When in 'Axial-Rotate' mode [disk-gizmo]:-

  Tab='Free-Rotate'
    
  Pick-3=Handle-Point
    OR
  Type 'Angle' + <Enter>
  
  Pick-4=Snap-Point
    OR
  Type 'Angle' + <Enter>
    
  If desired continue repeating Pick-3 & Pick-4 OR Type new 'Angle' etc...
    
  Enter/Double_Click=Confirm
    i.e. Use the current Rotation and Exit the tool.
  
  Esc/Exit=Cancel
    i.e. to Abort and Undo all changes at any stage...
    an 'Exit' occurs when you start to use another Tool.

When the tool 'exits' the Selection is ungrouped and remains highlighted.
The changes will 'Undo' in one step.
___________________________________________________________________________

Donations:
by PayPal to info @ revitrev.org
___________________________________________________________________________
Version:
1.1 20100825 First release.
1.2 20100826 Bug in empty selection trapped.
1.3 20100826 Encompassing 'globe' added.
2.0 20100828 New Ctrl/Alt/Tab Options for suspending & changing Pivot/Axes.
2.1 20100828 Glitch with Cursor colors during 'Tab' Options fixed.
             Changed the Options to Down_Arrow/Ctrl/Tab: 'Alt' is plain flaky!
2.2 20100828 Globe defaults to 150% of selections bounds.
             Tab-options rationalized.
2.3 20100829 Guide line added from Pivot to Axis Cursor during Tab-Option.
3.0 20109015 New 'Up-Arrow' option to rotate constrained around Axis added.
4.0 20100922 Radical overhaul of how the tool operates - see the notes...
             FreeRotate Toolbar added, and Zip file with FreeRotate folder.
4.1 20100922 Typo fixed with Axis on X_AXIS glitch.
             'Tab' press for MAC key code adjusted to 9, 15 and 48.
             Menu 'Windows > Instructor' help added...
             [This is not working - open ../FreeRotate/index.html in your 
             web-browser till this is fixed...]
4.2 20100923 Glitch with 'Tab' [9] on MAC etc trapped with key-up not key-down.
4.3 20100927 Typo in code corrected in line#708: was not affecting results.
             Gizmo Guides are displayed even when Guides are 'off' in Style.
4.4 20110307 Instructor now works...
___________________________________________________________________________
=end
#-----------------------------------------------------------------------------
require 'sketchup.rb'
#-----------------------------------------------------------------------------
class FreeRotate

def getExtents()
    bbox=Sketchup.active_model.bounds
    bbox.add(@ip.position)if @ip and @ip.valid?
    bbox.add(@ip1.position)if @ip1 and @ip1.valid?
    return bbox
end

# This is the standard Ruby initialize method that is called when you create
# a new object.
def initialize()
    @ip=nil
    @ip1=nil
    @ipp=nil
    @iph=nil
    $freerotategloberadius=1.5 if not $freerotategloberadius
end

# The activate method is called by SketchUp when the tool is first selected.
# it is a good place to put most of your initialization
def activate()
    # The Sketchup::InputPoint class is used to get 3D points from screen
    # positions.  It uses the SketchUp inferencing code.
    # In this tool, we will have two points for the endpoints of the line.
    @ip=Sketchup::InputPoint.new
    @ip1=Sketchup::InputPoint.new
    @ipp=Sketchup::InputPoint.new
    @iph=Sketchup::InputPoint.new
    @tline=nil
    @globe=[]
    @disk=[]
    @pivot=[]
    @p0=[]
    @axis=[]
    @handle=[]
    @snap=[]
    @group=nil
    @ctrl=false  ### bb center as pivot or axis
    @tab=true ### free / axial
    ###############################
    @model=Sketchup.active_model
    @view=@model.active_view
    @ents=@model.active_entities
    @sel=@model.selection
    @sela=[]
    if not @sel[0]
      UI.messagebox("Free Rotate: No Selection!")
      @state = -1
      self.onCancel(nil,@view)
      return nil
    end#if
    @sela=@sel.to_a
    Sketchup.active_model.start_operation("Free Rotate")
    @sel.clear
    ###
    @group=@ents.add_group(@sela)
    @tr=@group.transformation.to_a
    @bb=@group.bounds
    @center=@bb.center
    ###
    @msgPIVOTFREE="FREE ROTATE: Pick Pivot_Point OR Ctrl=BBox-Center, Tab=Axial_Rotate, Esc=Cancel..."
    @msgPIVOTAXIAL="AXIAL ROTATE: Pick Pivot_Point OR Ctrl=BBox-Center, Tab=Free_Rotate, Esc=Cancel..."
    @msgAXISFREE="FREE ROTATE: Pick Axis_Point OR Ctrl=BBox-Center, Tab=Axial_Rotate, Esc=Cancel..."
    @msgAXISAXIAL="AXIAL ROTATE: Pick Axis_Point OR Ctrl=BBox-Center, Tab=Free_Rotate, Esc=Cancel..."
    @msgFREE="FREE ROTATE: Left_Click/Enter/Double_Click=Confirm, Tab=Axial_Rotate, Esc=Cancel..."
    @msgAXIALHANDLE="AXIAL ROTATE: Pick Handle_Point OR Double_Click/Enter=Confirm, Type in an Angle, Tab=Free_Rotate, Esc=Cancel..."
    @msgAXIALSNAP="AXIAL ROTATE: Pick Snap_Point OR Double_Click/Enter=Confirm, Type in an Angle, Tab=Free_Rotate, Esc=Cancel..."
    ###
    @msg=@msgPIVOTFREE
    Sketchup::set_status_text(@msg)
    @state=0 ### Pivot-Point
    @ip.clear
    @ip1.clear
    @ipp.clear
    @iph.clear
    if @view
      @view.lock_inference
      @view.invalidate
    end
    @cgeom=@model.rendering_options["HideConstructionGeometry"]
    @model.rendering_options["HideConstructionGeometry"]=false
end

def make_globe()
    @globe.each{|e|e.erase! if e.valid?}if @globe
    @disk.each{|e|e.erase! if e.valid?}if @disk
    @globe << @ents.add_cpoint(@pivot)
    ###
    @rad=0.0
    (0..7).each{|i|
      dis=@pivot.distance(@bb.corner(i))
      @rad=dis if dis > @rad
    }
    @rad=@rad * $freerotategloberadius
    ### draw equator is now in great circles
    ### draw great circles
    yaxis=Y_AXIS.clone
    xaxis=X_AXIS.clone
    ang=0.0
    ###
    12.times{
      ps=@pivot.clone
      ps.offset!(yaxis,@rad)
      24.times{
        tr=Geom::Transformation.rotation(@pivot,xaxis,15.degrees)
        pe=ps.clone
        pe.transform!(tr)
        @globe << @ents.add_cline(ps,pe)
        @globe << @ents.add_cpoint(pe)
        ### part circle
        pp=ps.clone
        tr=Geom::Transformation.rotation(@pivot,Z_AXIS,15.degrees)
        pp.transform!(tr)
        @globe << @ents.add_cline(ps,pp)
        ###
        ps=pe.clone
      }
      ang=ang+15.degrees
      tr=Geom::Transformation.rotation(@pivot,Z_AXIS,15.degrees)
      yaxis.transform!(tr)
      xaxis.transform!(tr)
    }
    ### draw axial adii lines
    @globe << @ents.add_cline(@pivot,@pivot.clone.offset!(X_AXIS.clone,@rad))
    @globe << @ents.add_cline(@pivot,@pivot.clone.offset!(X_AXIS.clone.reverse,@rad))
    @globe << @ents.add_cline(@pivot,@pivot.clone.offset!(Y_AXIS.clone,@rad))
    @globe << @ents.add_cline(@pivot,@pivot.clone.offset!(Y_AXIS.clone.reverse,@rad))
    @globe << @ents.add_cline(@pivot,@pivot.clone.offset!(Z_AXIS.clone,@rad))
    @globe << @ents.add_cline(@pivot,@pivot.clone.offset!(Z_AXIS.clone.reverse,@rad))
end

def make_disk()
    @globe.each{|e|e.erase! if e.valid?}if @globe
    @disk.each{|e|e.erase! if e.valid?}if @disk
    @disk << @ents.add_cpoint(@pivot)
    ###
    @rad=0.0
    (0..7).each{|i|
      dis=@pivot.distance(@bb.corner(i))
      @rad=dis if dis > @rad
    }
    @rad=@rad * $freerotategloberadius / 2
    ### draw equator is now in great circles
    ### draw great circles
    @axis=@pivot.vector_to(@p0)
    @axis.normalize!
    ### first draw axis cline up/down
    @disk << @ents.add_cline(@pivot,@pivot.offset(@axis, 1.5*@rad))
    @disk << @ents.add_cline(@pivot,@pivot.offset(@axis, -1.5*@rad))
    if @axis==X_AXIS or @axis==X_AXIS.reverse
      yaxis=Geom::Vector3d.new(0,1,0) ### >>> Z_AXIS
    else
      yaxis=@axis.cross(X_AXIS)
    end#if
    ang=0.0
    r=0.0
    rad=@rad
    ###
    4.times{
     r=r+0.25
     rad=@rad*r
     12.times{
      ps=@pivot.clone
      ps.offset!(yaxis,rad) if yaxis.length!=0
      24.times{
        tr=Geom::Transformation.rotation(@pivot,@axis,15.degrees)
        pe=ps.clone
        pe.transform!(tr)
        @disk << @ents.add_cline(ps,pe)
        @disk << @ents.add_cpoint(pe)
        @disk << @ents.add_cline(@pivot,pe)
        ps=pe.clone
      }
      ang=ang+15.degrees
      tr=Geom::Transformation.rotation(@pivot,@axis,15.degrees)
      yaxis.transform!(tr)
     }
    }
    ###
end

# deactivate is called when the tool is deactivated because
# a different tool was selected etc
def deactivate(view)
    @model.rendering_options["HideConstructionGeometry"]=@cgeom
    self.undo_rotation()
    Sketchup::set_status_text("",SB_PROMPT)
    Sketchup::set_status_text("",SB_VCB_LABEL)
    Sketchup::set_status_text("",SB_VCB_VALUE)
    view.invalidate if view
    return nil
end

# onCancel is called when the user hits the escape key
def onCancel(flag, view)
    @model.rendering_options["HideConstructionGeometry"]=@cgeom
    self.undo_rotation()
    Sketchup::set_status_text("",SB_PROMPT)
    Sketchup::set_status_text("",SB_VCB_LABEL)
    Sketchup::set_status_text("",SB_VCB_VALUE)
    view.invalidate if view
    Sketchup.send_action("selectSelectionTool:")
    return nil
end

# The onMouseMove method is called whenever the user moves the mouse.
# because it is called so often, it is important to try to make it efficient.
# In a lot of tools, your main interaction will occur in this method.
def onMouseMove(flags, x, y, view)
    view.tooltip=@ip.tooltip
    if @state==0 ### pivot
      if @tab
        @msg=@msgPIVOTFREE
      else
        @msg=@msgPIVOTAXIAL
      end#if
      @ip.pick(view, x, y)
      view.tooltip="Pivot:"+@ip.tooltip
      if @ip != @ip1
        view.invalidate if @ip.display? or @ip1.display?
        @ip1.copy!(@ip)
      end
    elsif @state==1 ### axis
      if @tab
        @msg=@msgAXISFREE
      else
        @msg=@msgAXISAXIAL
      end#if
      @ip.pick(view, x, y)
      view.tooltip="Axis:"+@ip.tooltip
      if @ip != @ip1
        view.invalidate if @ip.display? or @ip1.display?
        @ip1.copy!(@ip)
        if @pivot != @ip1.position
          @p0r=@pivot.clone
          @p0r.offset!((@pivot.vector_to(@ip1.position)),1)
        end#if
      end#if
    elsif @state==2 ### handle
      @msg=@msgAXIALHANDLE
      @ip.pick(view, x, y)
      view.tooltip="Handle:"+@ip.tooltip
      if @ip != @ip1
        view.invalidate if @ip.display? or @ip1.display?
        @ip1.copy!(@ip)
      end
    elsif @state==3 ### snap
      @msg=@msgAXIALSNAP
      @ip.pick(view, x, y)
      view.tooltip="Snap:"+@ip.tooltip
      if @ip != @ip1
        view.invalidate if @ip.display? or @ip1.display?
        @ip1.copy!(@ip)
        if @pivot != @ip1.position and @handle != @ip1.position
          self.temp_axial_rotation()
        end#if
      end
    elsif @state==4
      @msg=@msgFREE
      @ip.pick(view, x, y)
      view.tooltip="Free:"+@ip.tooltip
      if @ip != @ip1
        view.invalidate if @ip.display? or @ip1.display?
        @ip1.copy!(@ip)
        if @pivot != @ip1.position
          @p0r=@pivot.clone
          @p0r.offset!((@pivot.vector_to(@ip1.position)),@rad)
          self.temp_free_rotation()
        end#if
      end
    end#if
    view.invalidate if @ip.display? or @ip1.display?
    Sketchup::set_status_text(@msg)
end

# The onLButtonUp method is called when the user releases the left mouse button.
def onLButtonDown(flags, x, y, view)
    @tline.erase! if @tline and @tline.valid?
    if @state==0 ### set pivot
      if @tab
        @msg=@msgAXISFREE
      else
        @msg=@msgAXISAXIAL
      end#if
      @pivot=@ip1.position
      @ipp.copy!(@ip1)
      @state=1
    elsif @state==1 ### set axis
      if @pivot != @ip1.position
        @p0=@ip1.position
        @msg=@msgFREE
        if @tab
          self.make_globe()
          @state=4
        else
          self.make_disk()
          @msg=@msgAXIALHANDLE
          @state=2
        end#if
        @p0r=@pivot.clone.offset(Z_AXIS,1)
      end#if 
    elsif @state==2 ### set handle
      #if @pivot != @ip1.position
        line=[@ip1.position,@axis]
        plane=[@pivot,@axis]### project onto plane
        int=Geom.intersect_line_plane(line,plane)
        return nil if int==@pivot
        @handle=int
        @iph.copy!(@ip1)
        @msg=@msgAXIALSNAP
        @state=3
      #end#if 
    elsif @state==3 ### set snap
      #if @handle != @ip1.position
        line=[@ip1.position,@axis]
        plane=[@pivot,@axis]### project onto plane
        int=Geom.intersect_line_plane(line,plane)
        return nil if int==@handle and int==@pivot
        @snap=int
        @msg=@msgAXIALHANDLE
        @state=2
      #end#if
    else ### ==4 commit
      self.commit_rotation()
    end#if
    view.invalidate if @ip.display? or @ip1.display?
    Sketchup::set_status_text(@msg)
end

def onReturn(view)
    self.commit_rotation()
end

def onLButtonDoubleClick(flags, x, y, view)
    self.commit_rotation() if not @tab
end

# onKeyDown is called when the user presses a key on the keyboard.
# We are checking it here to see if the user pressed the shift key
# so that we can do inference locking
def onKeyDown(key, repeat, flags, view)
    if key==CONSTRAIN_MODIFIER_KEY && repeat==1
        @shift_down_time=Time.now
        # if we already have an inference lock, then unlock it
        if view.inference_locked?
            # calling lock_inference with no arguments actually unlocks
            view.lock_inference
        elsif @state==0 && @ip1.valid?
            view.lock_inference(@ip1)
        elsif @state==1 && @ip1.valid?
            view.lock_inference(@ipp, @ip1)
        elsif @state==2 && @ip1.valid?
            view.lock_inference(@ip1)
        elsif @state==3 && @ip1.valid?
            view.lock_inference(@iph, @ip1)
        end
    end
    ###
    if key==VK_CONTROL ### center toggle
        if @state==0 ### @ pivot
          if @tab
            @msgAXISFREE="FREE ROTATE: Pick Axis_Point OR Tab=Axial_Rotate, Esc=Cancel..."
            @msg=@msgAXISFREE
          else
            @msgAXISAXIAL="AXIAL ROTATE: Pick Axis_Point OR Tab=Free_Rotate, Esc=Cancel..."
            @msg=@msgAXISAXIAL
          end#if
          @pivot=@center
          @state=1
        elsif @state==1 ### @ axis
          if @pivot != @center
            @p0=@center
            if @tab
              @msg=@msgAXIALHANDLE
              @state==4
            else ### ==1
              @msg=@msgFREE
              @state==2
            end#if
          end#if
        end#if
        view.invalidate if @ip.display? or @ip1.display?
    end#if
    ### 
    Sketchup::set_status_text(@msg)
    ###
end

# onKeyUp is called when the user releases the key
# We use this to unlock the interence
# If the user holds down the shift key for more than 1/2 second, then we
# unlock the inference on the release.  Otherwise, the user presses shift
# once to lock and a second time to unlock.
def onKeyUp(key, repeat, flags, view)
    if key==CONSTRAIN_MODIFIER_KEY && view.inference_locked? && (Time.now - @shift_down_time) > 0.5
        view.lock_inference
    end
    ###
    if key==9 or key==15 or key==48 ### TAB =9 for PC, MAC 15/48 for some locales ?
      if @tab
        self.make_disk()if @state>1
        if @state==0
          @msg=@msgAXIALPIVOT
        elsif @state==1
          @msg=@msgAXIALAXIS
        else
          @state=2
          @msg=@msgAXIALHANDLE
        end#if
        @tab=false ### axial
      else
        self.make_globe()if @state>1
        if @state==0
          @msg=@msgFREEPIVOT
        elsif @state==1
          @msg=@msgFREEAXIS
        else
          @state=4
          @msg=@msgFREE
        end#if
        @tab=true ### free
      end#if
      view.invalidate if @ip.display? or @ip1.display?
    end#if
    ###
    Sketchup::set_status_text(@msg)
    ###
end

def enableVCB?()
   return true
end

def onUserText(text, view)### use types in angle
    # We only accept input when the state is 1 (i.e. getting the second point)
    # This could be enhanced to also modify the last line created if a length
    # is entered after creating a line.
    return nil if @tab
    return nil if not @state>=2 and not @scale<=3
    # The user may type in something that we can't parse as a length
    # so we set up some exception handling to trap that
    begin
        value=text.to_f
    rescue
        # Error parsing the text
        UI.beep
        puts ("Cannot convert ")+text+(" to an Angle")
        value=nil
        Sketchup::set_status_text("",SB_VCB_VALUE)
    end
    return nil if not value or value=="" or value==0
    ang=value.degrees
    ang=360.degrees + ang if ang < 0
    if @handle and @handle.z
      snap=@handle.clone
    else
      if @axis==Z_AXIS or @axis==Z_AXIS.reverse
        vec=Geom::Vector3d.new(1,0,0)
      else
        vec=Geom::Vector3d.new(0,0,1)
      end#if
      @handle=@pivot.clone.offset!(@axis,1)
      vee=@axis.cross(vec)
      tr=Geom::Transformation.rotation(@pivot,vee,90.degrees)
      @handle.transform!(tr)
      snap=@handle.clone
    end#if
    tr=Geom::Transformation.rotation(@pivot,@axis,ang)
    snap.transform!(tr)
    @snap=snap.clone
    self.temp_axial_rotation(ang)
    @msg=@msgAXIALHANDLE
    Sketchup::set_status_text(@msg)
    @state=2
end

def resume(view=nil)
    view.invalidate if view
    Sketchup::set_status_text(@msg)
    #@model.pop_tool(@model.tools.push_tool(@model.active_tool_id))
end

# The draw method is called whenever the view is refreshed.  
# It lets the tool draw any temporary geometry that it needs to.
def draw(view)
  @tline.erase! if @tline and @tline.valid?
  if @state==0 ### pivot
    if @ip.valid? or @ip1.valid? or @ip.display? or @ip1.display?
      view.line_width=3
      view.draw_points([@ip1.position],9,1)
      view.invalidate
    end#if
  elsif @state==1 ### axis
    if @ip.valid? or @ip1.valid?
      view.line_width=3
      view.draw_points([@ip1.position],9,1)
      view.draw_points([@pivot],9,1,"Black")
      view.set_color_from_line(@pivot, @ip1.position)
      view.draw_line(@pivot, @ip1.position)
      @tline=@ents.add_cline(@pivot, @ip1.position)
    end#if
  elsif @state==4 ### rotating
    if @ip.valid? or @ip1.valid?
      view.line_width=3
      view.draw_points([@pivot],9,1,"Black")
      view.draw_points([@p0r],9,1,"Magenta")
      if @state==0
        view.line_width=1
        @ip1.draw(view) if @ip.display? or @ip1.display?
        view.set_color_from_line(@pivot, @ip1.position)
        view.draw_points([@ip1.position],9,1)
        view.draw_line(@pivot, @ip1.position)
        @tline=@ents.add_cline(@pivot, @ip1.position)
      end
    end
  elsif @state==2 ### constrained axial rotation handle
    view.line_width=3
    @ip1.draw(view) if @ip.display? or @ip1.display?
    view.draw_points([@ip1.position],9,1)
    view.draw_points([@pivot],9,1,"DarkOrange")
  elsif @state==3 ### constrained axial rotation snap
    view.line_width=3
    @ip1.draw(view) if @ip.display? or @ip1.display?
    view.draw_points([@ip1.position],9,1)
    view.draw_points([@pivot],9,1,"Magenta")
  end#if
end

###############################################

def commit_rotation()
    @model.rendering_options["HideConstructionGeometry"]=@cgeom
    @state=5
    @tline.erase! if @tline and @tline.valid?
    @globe.each{|e|e.erase! if e.valid?}if @globe
    @disk.each{|e|e.erase! if e.valid?}if @disk
    @group.explode if @group and @group.valid?
    self.reselect()
    Sketchup.active_model.commit_operation
    Sketchup.send_action("selectSelectionTool:")
    return nil
end

def temp_free_rotation()
  return nil if @state != 4
  p0=@p0.clone
  p1=@ip1.position
  if @pivot != p0 and @pivot != p1 and p0 != p1
    vec_p_0=@pivot.vector_to(p0)
    vec_p_1=@pivot.vector_to(p1)
    axis=vec_p_0.cross(vec_p_1)
    return nil if axis.length==0
    ang=vec_p_0.angle_between(vec_p_1)
    return nil if ang==0
    tr=Geom::Transformation.rotation(@pivot,axis,ang)
    @group.transform!(tr)
  end
  @p0=p1.clone
end

def temp_axial_rotation(angin=nil)
  return nil if not @state>=2
  p0=@handle.clone
  line=[@ip1.position,@axis]
  plane=[@pivot,@axis]
  p1=Geom.intersect_line_plane(line,plane)
  p1=@snap if angin
  if @pivot != p0 and @pivot != p1 and p0 != p1
    vec_p_0=@pivot.vector_to(p0)
    vec_p_1=@pivot.vector_to(p1)
    axis=vec_p_0.cross(vec_p_1)
    return nil if axis.length==0 and not angin
    ang=vec_p_0.angle_between(vec_p_1)
    return nil if ang==0
    axis=@axis if angin
    ang=angin if angin
    tr=Geom::Transformation.rotation(@pivot,axis,ang)
    @group.transform!(tr)
  end
  @handle=p1.clone    
end

def undo_rotation()
    @model.rendering_options["HideConstructionGeometry"]=@cgeom
    @state=4
    @tline.erase! if @tline and @tline.valid?
    @globe.each{|e|e.erase! if e.valid?}if @globe
    @disk.each{|e|e.erase! if e.valid?}if @disk
    @group.transformation=@tr if @group and @group.valid?
    @group.explode if @group and @group.valid?
    Sketchup.active_model.commit_operation
    self.reselect()
    return nil
end

def reselect()
    @sela.each{|e|@sel.add(e)if e.valid?}if @sela
end

def getInstructorContentDirectory()
   path=File.expand_path(File.join(File.dirname(__FILE__),"FreeRotate")).split('/')
   0.upto(path.length-3){|i| path[i]='..' }
   if [PLATFORM].grep(/mswin/)==[PLATFORM] ### PC
     return path.join('/')+'/'
   else
     return '/'+path.join('/')+'/'
   end#if
end

end # class

#----------------------------------------------------
# shortcut
def freerotate()
    Sketchup.active_model.select_tool(FreeRotate.new)
end
######-----------------------------------------------

### add menu item etc #####################################################
if not file_loaded?(File.basename(__FILE__))
  ### make command1
  text10 = "Free Rotate"
  text11 = "Free Rotate"
  text12 = "Free-Rotate / Axial-Rotate a Selection about a Point"+"..."
  cmd1=UI::Command.new(text10){Sketchup.active_model.select_tool(FreeRotate.new)}
  cmd1.tooltip=text11
  cmd1.status_bar_text=text12
  UI.menu("Plugins").add_item(cmd1)
  ### toolbar
  cmd1.small_icon=File.join(File.dirname(__FILE__), "FreeRotate/FreeRotate16x16.png")
  cmd1.large_icon=File.join(File.dirname(__FILE__), "FreeRotate/FreeRotate24x24.png")
  toolbar=UI::Toolbar.new(text10)
  toolbar.add_item(cmd1)
  toolbar.restore if toolbar.get_last_state==TB_VISIBLE
  ###
end#if
file_loaded(File.basename(__FILE__))
###
